#define UMAD(a, b, c)  ( (a) * (b) + (c) )
#define M_PI 3.14159265f

typedef unsigned int uint;

typedef struct{
    float x;
    float y;
    float z;
} Vector3;

// SIMULATION PARAMETERS
typedef struct {    
    Vector3 gravity;
    float particleRadius;
	float particleMass;
	float boundaryDamping;
	float globalDamping;

	float smoothingLength;
	float viscosity;

	Vector3 worldOrigin;
	Vector3 worldSize;
	float cellSize;

	Vector3 gridSize;
	int numGridCells;

	float mu0;
	float rho0;
	float kv;
	float stiffness;
	float dv;
} simParams_t;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MATRIX OPERATIONS
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef struct {
	float m11, m12, m13;
	float m21, m22, m23;
	float m31, m32, m33;
} Matrix3x3;

Matrix3x3 newMatrix(float x) {
	Matrix3x3 result;
	result.m11 = 0.0f;
	result.m12 = 0.0f;
	result.m13 = 0.0f;
	result.m21 = 0.0f;
	result.m22 = 0.0f;
	result.m23 = 0.0f;
	result.m31 = 0.0f;
	result.m32 = 0.0f;
	result.m33 = 0.0f;
	return result;
}

Matrix3x3 outer(float4 u, float4 v) {
	Matrix3x3 result;
	result.m11 = u.x * v.x;
	result.m12 = u.x * v.y;
	result.m13 = u.x * v.z;
	result.m21 = u.y * v.x;
	result.m22 = u.y * v.y;
	result.m23 = u.y * v.z;
	result.m31 = u.z * v.x;
	result.m32 = u.z * v.y;
	result.m33 = u.z * v.z;
	return result;
}

Matrix3x3 add(Matrix3x3 m1, Matrix3x3 m2) {
	Matrix3x3 result;
	result.m11 = m1.m11 + m2.m11;
	result.m12 = m1.m12 + m2.m12;
	result.m13 = m1.m13 + m2.m13;
	result.m21 = m1.m21 + m2.m21;
	result.m22 = m1.m22 + m2.m22;
	result.m23 = m1.m23 + m2.m23;
	result.m31 = m1.m31 + m2.m31;
	result.m32 = m1.m32 + m2.m32;
	result.m33 = m1.m33 + m2.m33;
	return result;
}

float4 product(Matrix3x3 m, float4 u) {
	float4 result = (float4)(0,0,0,0);
	result.x = m.m11*u.x + m.m12*u.y + m.m13*u.z;
	result.y = m.m21*u.x + m.m22*u.y + m.m23*u.z;
	result.z = m.m31*u.x + m.m32*u.y + m.m33*u.z;
	return result;
}

Matrix3x3 mult(Matrix3x3 m, float x) {
	Matrix3x3 result;
	result.m11 = m.m11 * x;
	result.m12 = m.m12 * x;
	result.m13 = m.m13 * x;
	result.m21 = m.m21 * x;
	result.m22 = m.m22 * x;
	result.m23 = m.m23 * x;
	result.m31 = m.m31 * x;
	result.m32 = m.m32 * x;
	result.m33 = m.m33 * x;
	return result;
}

Matrix3x3 transpose(Matrix3x3 m) {
	Matrix3x3 result;
	result.m11 = m.m11;
	result.m12 = m.m21;
	result.m13 = m.m31;
	result.m21 = m.m12;
	result.m22 = m.m22;
	result.m23 = m.m32;
	result.m31 = m.m13;
	result.m32 = m.m23;
	result.m33 = m.m33;
	return result;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// /END MATRIX OPERATIONS
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// GET THE COORDINATES OF THE CELL FROM THE POSITON OF A PARTICLE
int4 getCell(float4 pos, __constant simParams_t* g_params) {
	int4 gp;
	gp.x = (int)floor((pos.x - g_params->worldOrigin.x) / g_params->cellSize);
	gp.y = (int)floor((pos.y - g_params->worldOrigin.y) / g_params->cellSize);
	gp.z = (int)floor((pos.z - g_params->worldOrigin.z) / g_params->cellSize);
	gp.w = 0;
	return gp;
	
}

// GET THE HASH VALUE ASSOCIATED WITH A CELL
uint getHashValue(int4 cell, __constant simParams_t* g_params) {
	cell.x = cell.x & ((int)g_params->gridSize.x-1);
	cell.y = cell.y & ((int)g_params->gridSize.y-1);
	cell.z = cell.z & ((int)g_params->gridSize.z-1);
	return UMAD( UMAD(cell.z, g_params->gridSize.y, cell.y), g_params->gridSize.x, cell.x );
	//return cell.x + cell.y * (int)g_params->gridSize.x + cell.z * (int)g_params->gridSize.x * (int)g_params->gridSize.y;
}

// CALC HASH VALUES FOR EVERY PARTICLE
__kernel void calcHash (
	__global float4* g_pos,
	__global uint* g_hash,
	__global uint* g_index,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);

	if ( index > numParticles )
		return;

	float4 pos = g_pos[index];
	int4 cell = getCell(pos, g_params);
	uint hash = getHashValue(cell, g_params);

	g_hash[index] = hash;
	g_index[index] = index;
}

// FIND THE PARTICLE INDICES OF THE FIRST AND LAST PARTICLES IN EACH CELL
__kernel void findCellBounds (
	__global const float4* g_pos,
	__global const float4* g_vel,
	__global float4* g_sortedPos,
	__global float4* g_sortedVel,
	__global const uint* g_index,
	__global const uint* g_hash,
	__global uint* g_cellStart,
	__global uint* g_cellEnd,
	uint numParticles
) {
	const uint index = get_global_id(0);
	if ( index > numParticles )
		return;

	uint hash = g_hash[index];

	if ( index == 0 )
		g_cellStart[hash] = 0;
	else {
		uint prevHash = g_hash[index-1];
		if ( hash != prevHash )
			g_cellEnd[prevHash] = g_cellStart[hash] = index;
	};
	if ( index == numParticles - 1 )
		g_cellEnd[hash] = numParticles;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SMOOTHING FUNCTIONS
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

float wPoly6(float4 relPos, float r, float h) {
	if ( r > h ) return 0.0f;
	return (315/(64*M_PI*h*h*h*h*h*h*h*h*h)) * (h*h-r*r) * (h*h-r*r) * (h*h-r*r);
}

float4 grad_wPoly6(float4 relPos, float r, float h) {
	if ( r > h ) return (float4)(0,0,0,0);
	return -1 * (float4)(relPos.x / r, relPos.y / r, relPos.z / r, 0) * (945/(32*M_PI*h*h*h*h*h*h*h*h*h)) * (h*h-r*r) * (h*h-r*r);
}

float lap_wPoly6(float4 relPos, float r, float h) {
	if ( r > h || r == 0.0f ) return 0.0f;
	return -1 * (945/(32*M_PI*h*h*h*h*h*h*h*h*h)) * (h*h - r*r) * (3*h*h - 7*r*r);
}

float4 grad_wSpiky(float4 relPos, float r, float h) {
	if ( r > h ) return (float4)(0,0,0,0);
	return -1 * (float4)(relPos.x / r, relPos.y / r, relPos.z / r, 0) * (45/(M_PI*h*h*h*h*h*h)) * (h-r) * (h-r);
}

float lap_wVis(float4 relPos, float r, float h) {
	if ( r > h ) return 0.0f;
	return (45/(M_PI*h*h*h*h*h*h)) * (h-r);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// /END SMOOTHING FUNCTIONS
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// INITIALIZE DENSITY FOR EACH PARTICLE
__kernel void initializeDensity (
	__global const float4* g_pos,
	__global float4* g_initden,
	float m_snow,
	float m_mixture,
	__global const uint* g_cellStart,
	__global const uint* g_cellEnd,
	__global const uint* g_index,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);
	if ( index > numParticles )
		return;

	float4 pos_i = g_pos[g_index[index]];

	float den_snow = 0.0f;
	float den_mixture = 0.0f;

	int4 cell = getCell(pos_i, g_params);

	for ( int z = -1; z <= 1; z++ ) {
		for ( int y = -1; y <= 1; y++ ) {
			for ( int x = -1; x <= 1; x++ ) {
				int4 nbCell = cell + (int4)(x,y,z,0);
				uint hash = getHashValue(nbCell, g_params);

				uint startIndex = g_cellStart[hash];

				if ( startIndex == 0xFFFFFFFFU )
					continue;

				uint endIndex = g_cellEnd[hash];

				for ( uint j = startIndex; j < endIndex; ++j ) {
					if ( j == index )
						continue;

					float4 pos_j = g_pos[g_index[j]];
					float4 x_ij = (float4)(pos_i.x - pos_j.x, pos_i.y - pos_j.y, pos_i.z - pos_j.z, 0);

					float r2 = x_ij.x * x_ij.x + x_ij.y * x_ij.y + x_ij.z * x_ij.z;
					float r = sqrt(r2);
					float h = g_params->smoothingLength;

					float wpoly6 = wPoly6(x_ij,r,h);

					den_snow += m_snow * wpoly6;
					den_mixture += m_mixture * wpoly6;
				}
			}
		}
	}

	float4 initden_i = g_initden[g_index[index]];
	initden_i.x = den_snow;
	initden_i.y = den_mixture;
	g_initden[g_index[index]] = initden_i;

}

// COMPUTE DENSITY FOR EACH PARTICLE
__kernel void computeDensity (
	__global const float4* g_pos,
	__global const float4* g_vel,
	__global float4* g_svar,
	__global float4* g_pvar,
	__global const uint* g_cellStart,
	__global const uint* g_cellEnd,
	__global const uint* g_index,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);

	if ( index > numParticles )
		return;

	// POSITION AND VELOCITY
	float4 pos_i = g_pos[g_index[index]];
	float4 vel_i = g_vel[g_index[index]];

	float4 svar_i = g_svar[g_index[index]];
	float4 pvar_i = g_pvar[g_index[index]];

	float den_i = svar_i.x;
	float den_ref = svar_i.y;
	float den_rest = svar_i.z;

	// VARIABLES TO BE CALCULATED
	float num = 0.0f;
	float den = 0.0f;
	//float rho = 0.0f;

	// GET PARTICLE CELL
	int4 cell = getCell(pos_i, g_params);

	for ( int z = -1; z <= 1; z++ ) {
		for ( int y = -1; y <= 1; y++ ) {
			for ( int x = -1; x <= 1; x++ ) {
				int4 nbCell = cell + (int4)(x,y,z,0);
				uint hash = getHashValue(nbCell, g_params);

				uint startIndex = g_cellStart[hash];

				if ( startIndex == 0xFFFFFFFFU )
					continue;

				uint endIndex = g_cellEnd[hash];

				for ( uint j = startIndex; j < endIndex; ++j ) {
					if ( j == index )
						continue;

					// POSITION AND VELOCITY
					float4 pos_j = g_pos[g_index[j]];
					float4 vel_j = g_vel[g_index[j]];

					// RELATIVE POSITION AND VELOCITIES
					float4 x_ij = (float4)(pos_i.x - pos_j.x, pos_i.y - pos_j.y, pos_i.z - pos_j.z, 0);

					float4 pvar_j = g_pvar[g_index[j]];
					float m_j = pvar_j.w;

					float4 svar_j = g_svar[g_index[j]];
					float den_j = svar_j.x;

					float r2 = x_ij.x * x_ij.x + x_ij.y * x_ij.y + x_ij.z * x_ij.z;
					float r = sqrt(r2);
					float h = g_params->smoothingLength;

					float wpoly6 = wPoly6(x_ij,r,h);

					num += m_j * wpoly6;
					den += (m_j/den_j) * wpoly6;
					//rho += m_j * wpoly6;
				}
			}
		}
	}

	float rho = num / den;

	float den_difference = den_rest - den_ref;
	float density_i = rho - den_difference;

	svar_i.x = density_i;

	g_svar[g_index[index]] = svar_i;
}

// COMPUTE PRESSURE FOR EACH PARTICLE
__kernel void computePressure (
	__global const float4* g_pos,
	__global const float4* g_vel,
	__global float4* g_svar,
	__global float4* g_pvar,
	__global const uint* g_cellStart,
	__global const uint* g_cellEnd,
	__global const uint* g_index,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);

	if ( index > numParticles )
		return;

	// POSITION AND VELOCITY
	float4 pos_i = g_pos[g_index[index]];
	float4 vel_i = g_vel[g_index[index]];

	// VARIABLES
	float4 svar_i = g_svar[g_index[index]];
	float4 pvar_i = g_pvar[g_index[index]];

	float den_i = svar_i.x;
	float logden_i = log10(den_i);
	float den_ref = svar_i.y;
	float den_rest = svar_i.z;

	float k = g_params->stiffness;

	// VARIABLES TO BE CALCULATED
	float lap_logrho = 0.0f;

	// GET PARTICLE CELL
	int4 cell = getCell(pos_i, g_params);

	for ( int z = -1; z <= 1; z++ ) {
		for ( int y = -1; y <= 1; y++ ) {
			for ( int x = -1; x <= 1; x++ ) {
				int4 nbCell = cell + (int4)(x,y,z,0);
				uint hash = getHashValue(nbCell, g_params);

				uint startIndex = g_cellStart[hash];

				if ( startIndex == 0xFFFFFFFFU )
					continue;

				uint endIndex = g_cellEnd[hash];

				for ( uint j = startIndex; j < endIndex; ++j ) {
					if ( j == index )
						continue;

					// POSITION AND VELOCITY
					float4 pos_j = g_pos[g_index[j]];
					float4 vel_j = g_vel[g_index[j]];

					// RELATIVE POSITION AND VELOCITIES
					float4 x_ij = (float4)(pos_i.x - pos_j.x, pos_i.y - pos_j.y, pos_i.z - pos_j.z, 0);
					float4 v_ij = (float4)(vel_i.x - vel_j.x, vel_i.y - vel_j.y, vel_i.z - vel_j.z, 0);

					float4 svar_j = g_svar[g_index[j]];
					float4 pvar_j = g_pvar[g_index[j]];
					
					float den_j = svar_j.x;
					float logden_j = log10(den_j);
					float m_j = pvar_j.w;

					float r2 = x_ij.x * x_ij.x + x_ij.y * x_ij.y + x_ij.z * x_ij.z;
					float r = sqrt(r2);
					float h = g_params->smoothingLength;

					float wpoly6 = wPoly6(x_ij, r, h);
					float4 grad_wpoly6 = grad_wPoly6(x_ij, r, h);
					float lap_wpoly6 = lap_wPoly6(x_ij, r, h);

					lap_logrho += m_j * ( (logden_i/(den_i*den_i)) + (logden_j/(den_j*den_j)) ) * lap_wpoly6;
				}
			}
		}
	}

	lap_logrho *= den_i;

	float epsilon = lap_logrho;

	pvar_i.z = epsilon;

	float pressure_i = ( (k * den_ref) / 7 ) * ( pow( (den_i/den_ref), 7 ) - 1);
	//float pressure_i = k * ( den_i - den_ref );
	svar_i.w = pressure_i;

	g_svar[g_index[index]] = svar_i;
	g_pvar[g_index[index]] = pvar_i;
}

// COMPUTE FORCES FOR EACH PARTICLE
__kernel void computeForces (
	__global const float4* g_pos,
	__global const float4* g_vel,
	__global float4* g_svar,
	__global float4* g_pvar,
	__global float4* g_force,
	__global float4* g_color,
	__global const uint* g_cellStart,
	__global const uint* g_cellEnd,
	__global const uint* g_index,
	float dt,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);

	if ( index > numParticles )
		return;

	// POSITION AND VELOCITY
	float4 pos_i = g_pos[g_index[index]];
	float4 vel_i = g_vel[g_index[index]];

	// VARIABLES
	float4 svar_i = g_svar[g_index[index]];
	float4 pvar_i = g_pvar[g_index[index]];

	float den_i = svar_i.x;
	float logden_i = log10(den_i);
	float den_ref = svar_i.y;
	float den_rest = svar_i.z;
	float p_i = svar_i.w;
	float epsilon_i = pvar_i.z;

	float kv = g_params->kv;
	float mu0 = g_params->mu0;

	// VARIABLES TO BE CALCULATED
	float4 grad_p = (float4)(0,0,0,0);
	//float4 grad_alpha = (float4)(0,0,0,0);
	float4 grad_rho = (float4)(0,0,0,0);
	//float4 curl_v = (float4)(0,0,0,0);
	//float4 grad_logrho = (float4)(0,0,0,0);
	//float4 grad_beta = (float4)(0,0,0,0);
	float4 lap_v = (float4)(0,0,0,0);
	float4 grad_epsilon = (float4)(0,0,0,0);
	float lap_rho = 0.0f;
	Matrix3x3 grad_v = newMatrix(0.0f);

	// GET PARTICLE CELL
	int4 cell = getCell(pos_i, g_params);

	for ( int z = -1; z <= 1; z++ ) {
		for ( int y = -1; y <= 1; y++ ) {
			for ( int x = -1; x <= 1; x++ ) {
				int4 nbCell = cell + (int4)(x,y,z,0);
				uint hash = getHashValue(nbCell, g_params);

				uint startIndex = g_cellStart[hash];

				if ( startIndex == 0xFFFFFFFFU )
					continue;

				uint endIndex = g_cellEnd[hash];

				for ( uint j = startIndex; j < endIndex; ++j ) {
					if ( j == index )
						continue;

					// POSITION AND VELOCITY
					float4 pos_j = g_pos[g_index[j]];
					float4 vel_j = g_vel[g_index[j]];

					// RELATIVE POSITION AND VELOCITIES
					float4 x_ij = (float4)(pos_i.x - pos_j.x, pos_i.y - pos_j.y, pos_i.z - pos_j.z, 0);
					float4 v_ij = (float4)(vel_i.x - vel_j.x, vel_i.y - vel_j.y, vel_i.z - vel_j.z, 0);
					float4 v_ji = (float4)(vel_j.x - vel_i.x, vel_j.y - vel_i.y, vel_j.z - vel_i.z, 0);

					float4 svar_j = g_svar[g_index[j]];
					float4 pvar_j = g_pvar[g_index[j]];
					
					float den_j = svar_j.x;
					float logden_j = log10(den_j);
					float m_j = pvar_j.w;

					float p_j = svar_j.w;
					float epsilon_j = pvar_j.z;

					float r2 = x_ij.x * x_ij.x + x_ij.y * x_ij.y + x_ij.z * x_ij.z;
					float r = sqrt(r2);
					float h = g_params->smoothingLength;

					float4 grad_wpoly6 = grad_wPoly6(x_ij, r, h);
					float lap_wpoly6 = lap_wPoly6(x_ij, r, h);
					float4 grad_wspiky = grad_wSpiky(x_ij, r, h);
					float lap_wvis = lap_wVis(x_ij, r, h);

					grad_p += m_j * ( (p_i/(den_i*den_i)) + (p_j/(den_j*den_j)) ) * grad_wspiky;
					//grad_alpha += m_j * ( (alpha_i/(den_i*den_i)) + (alpha_j/(den_j*den_j)) ) * grad_wpoly6;
					grad_rho += m_j * ( (1/den_i) + (1/den_j) ) * grad_wpoly6;
					//curl_v += m_j * cross(v_ij, grad_wpoly6);
					//grad_logrho += m_j * ( (logden_i/(den_i*den_i)) + (logden_j/(den_j*den_j)) ) * grad_wpoly6;
					//grad_beta += m_j * ( (beta_i/(den_i*den_i)) + (beta_j/(den_j*den_j)) ) * grad_wpoly6;
					lap_v += m_j * v_ji * lap_wvis;
					grad_epsilon += m_j * ( (epsilon_i/(den_i*den_i)) + (epsilon_j/(den_j*den_j)) ) * grad_wpoly6;
					lap_rho += m_j * ( (1/den_i) + (1/den_j) ) * lap_wpoly6;
					grad_v = add(grad_v, mult( outer(v_ji, grad_wpoly6), m_j ));

				}
			}
		}
	}
	grad_p *= den_i;
	//grad_alpha *= den_i;
	grad_rho *= den_i;
	//curl_v /= den_i;
	//grad_logrho *= den_i;
	//grad_beta *= den_i;
	lap_v /= den_i;
	grad_epsilon *= den_i;
	lap_rho *= den_i;
	grad_v = mult( grad_v, 1/den_i );

	float4 f_a = -1 * grad_p;
	//float4 f_b = -2 * kv * grad_alpha;
	//float4 f_c = (2/den_i) * kv * alpha_i * grad_rho;
	//float4 f_d = 2 * kv * den_i * cross(curl_v, grad_logrho);
	//float4 f_e = 2 * kv * den_i * grad_beta;
	float4 f_f = g_params->viscosity * lap_v;
	float4 f_g = -4 * kv * mu0 * grad_epsilon;

	float4 F_b = -2 * kv * product( transpose(grad_v), grad_rho );
	float4 F_c = 2 * kv * product( grad_v, grad_rho );

	//float4 forces = f_a + f_b + f_c + f_d + f_e + f_f + f_g;
	float4 forces = f_a + F_b + F_c + f_f + f_g; 

	float delta_rho = (2 * kv * lap_rho) * dt;
	den_ref += delta_rho;
	svar_i.y = den_ref;

	g_svar[g_index[index]] = svar_i;

	g_force[g_index[index]] = forces;
	g_color[g_index[index]] = (float4)(1.0, 1.0, 1.0f, 1.0f);
	
}

__kernel void integrate (
	__global float4* g_pos,
	__global float4* g_vel,
	__global float4* g_svar,
	__global float4* g_force,
	__global const uint* g_index,
	float dt,
	uint numParticles,
	__constant simParams_t* g_params
) {
	const uint index = get_global_id(0);

	if ( index > numParticles )
		return;

	float4 pos = g_pos[g_index[index]];
	float4 vel = g_vel[g_index[index]];

	pos.w = 1.0f;
	vel.w = 0.0f;

	float4 svar_i = g_svar[g_index[index]];
	float den_i = svar_i.x;

	float4 force = g_force[g_index[index]];
	
	float4 acc = force / den_i;

	acc += (float4)(g_params->gravity.x, g_params->gravity.y, g_params->gravity.z, 0);

	vel += acc * dt;
	pos += vel * dt;

	if ( pos.x < -1.0f + g_params->particleRadius ) {
		pos.x = -1.0f + g_params->particleRadius;
		vel.x *= 0.0f; //g_params->boundaryDamping;
	}
	if ( pos.x > 9.0f - g_params->particleRadius ) {
		pos.x = 9.0f - g_params->particleRadius;
		vel.x *= g_params->boundaryDamping;
	}
	if ( pos.y < -1.0f + g_params->particleRadius ) {
		pos.y = -1.0f + g_params->particleRadius;
		vel *= 0.0f; // (float4)(0,0,0,0); //
	}
	if ( pos.y > 3.0f - g_params->particleRadius ) {
		pos.y = 3.0f - g_params->particleRadius;
		vel.y *= g_params->boundaryDamping;
	}
	if ( pos.z > 2.0f - g_params->particleRadius ) {
		pos.z = 2.0f - g_params->particleRadius;
		vel.z *= g_params->boundaryDamping;
	}
	if ( pos.z < -2.0f + g_params->particleRadius ) {
		pos.z = -2.0f + g_params->particleRadius;
		vel.z *= g_params->boundaryDamping;
	}

	g_pos[g_index[index]] = pos;
	g_vel[g_index[index]] = vel;
}
