#include "clParticles.h"

using namespace cl;

clParticles* g_psystem = 0;

clParticles::clParticles(uint numParticles, float gridSize, float particleRadius) :
m_numParticles(numParticles),
m_initialized(false),
m_fileName("../Snowsim.cl"),
cVel(0),
cPos(0),
cSortedPos(0),
cSortedVel(0),
cSimVar(0),
cParVar(0),
cForce(0),
cInitDen(0),
cCellStart(0),
cCellEnd(0),
cIndex(0),
cHash(0)
{
	m_params.particleRadius = particleRadius;
	m_params.gravity = Vector3(88888.11f, -44444.34f, 0.0f);
	m_params.boundaryDamping = 1.0f;
	m_params.globalDamping = 1.0f;
	
	m_params.worldSize = Vector3(10.0, 4.0, 4.0);

	m_params.gridSize = Vector3(gridSize*5, gridSize*2, gridSize*2);
	m_params.numGridCells = (int)(m_params.gridSize.x * m_params.gridSize.y * m_params.gridSize.z);

	m_params.cellSize = 2.0 / 64.0;

	m_params.smoothingLength = m_params.cellSize * 1.5f;

	m_params.worldOrigin = Vector3(-1.0f, -1.0f, -2.0f);

	printf("World Size: %f x %f x %f\n", m_params.worldSize.x, m_params.worldSize.y, m_params.worldSize.z);
	printf("Cell Size: %f\n", m_params.cellSize);
	printf("GridSize: %f x %f x %f\n", m_params.gridSize.x, m_params.gridSize.y, m_params.gridSize.z);
	printf("Number of Grid Cells: %d\n", m_params.numGridCells);
	printf("Smoothing Length: %f\n", m_params.smoothingLength);

}

clParticles::~clParticles(void)
{
	close();
	m_numParticles = 0;
}

static std::string readFromFile(const char* file) {
	std::ifstream sourceFile(file);
	std::string sourceCode(std::istreambuf_iterator<char>(sourceFile), (std::istreambuf_iterator<char>()));

	return sourceCode;
}

void
clParticles::createProgram() {
	std::string sourceCode = readFromFile(m_fileName);
	Program::Sources source(1, std::make_pair(sourceCode.c_str(), sourceCode.length()+1));

	m_clParticlesProgram = Program(g_clAdmin->getContext(), source);
	m_clParticlesProgram.build(g_clAdmin->getDevices());
}

void 
clParticles::init() {
	try {
		// ALLOCATE HOST MEMORY
		cPos = new float[4*m_numParticles];
		cVel = new float[4*m_numParticles];
		cSortedPos = new float[4*m_numParticles];
		cSortedVel = new float[4*m_numParticles];
		cSimVar = new float[4*m_numParticles];
		cParVar = new float[4*m_numParticles];
		cHash = new uint[m_numParticles];
		cIndex = new uint[m_numParticles];
		cCellStart = new uint[m_params.numGridCells];
		cCellEnd = new uint[m_params.numGridCells];
		cForce = new float[4*m_numParticles];
		cInitDen = new float[4*m_numParticles];

		// SORTER 
		m_sorter = new clSort("../BitonicSort_b.cl");
		createProgram();

		// DECLARE KERNELS
		kCalcHash = Kernel(m_clParticlesProgram, "calcHash");
		kFindCellBounds = Kernel(m_clParticlesProgram, "findCellBounds");
		kInitializeDensity = Kernel(m_clParticlesProgram, "initializeDensity");
		kComputeDensity = Kernel(m_clParticlesProgram, "computeDensity");
		kComputePressure = Kernel(m_clParticlesProgram, "computePressure");
		kComputeForces = Kernel(m_clParticlesProgram, "computeForces");
		kIntegrate = Kernel(m_clParticlesProgram, "integrate");

		// ALLOCATE DEVICE MEMORY
		bVel = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bSortedVel = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bSortedPos = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bSimVar = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bParVar = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bCellStart = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_params.numGridCells * sizeof(uint));
		bCellEnd = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_params.numGridCells * sizeof(uint));
		bIndex = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(uint));
		bHash = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(uint));
		bParams = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, sizeof(simParams_t));
		bForce = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));
		bInitDen = Buffer(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_numParticles * sizeof(Vector4));

		// Allocate VBO storage
		glGenBuffers(1, &m_posVBO);	
		glBindBuffer(GL_ARRAY_BUFFER, m_posVBO);
		glBufferData(GL_ARRAY_BUFFER, m_numParticles * sizeof(Vector4), 0, GL_DYNAMIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glFinish();

		glGenBuffers(1, &m_colorVBO);	
		glBindBuffer(GL_ARRAY_BUFFER, m_colorVBO);
		glBufferData(GL_ARRAY_BUFFER, m_numParticles * sizeof(Vector4), 0, GL_DYNAMIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glFinish();
		
		bPos = BufferGL(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_posVBO);
		bColor = BufferGL(g_clAdmin->getContext(), CL_MEM_READ_WRITE, m_colorVBO);
		m_clVector.push_back(bPos);
		m_clVector.push_back(bColor);
		m_initialized = true;

		// CALL FUNCTION TO INITIALIZE PHYSICAL PARAMTERS
		reset(CONFIG_GRID);

		copyDataToGPU(&m_params, bParams, sizeof(simParams_t));

	} catch(Error error) {
		std::cout << "Error in init()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}

void 
clParticles::close() {

}

void
clParticles::sort() {
 	m_sorter->bitonicSort(bHash, bIndex, bHash, bIndex);
}

//*********************

void
clParticles::update() { // RUNS SIMULATION
	try {
		assert(m_initialized);

		glFinish();

		g_clAdmin->getCommandQueue().enqueueAcquireGLObjects(&m_clVector);
 
		calcHash();
		sort();
		findCellBounds();
		computeDensity();
		computePressure();
		computeForces();
		integrate();	

		g_clAdmin->getCommandQueue().enqueueReleaseGLObjects(&m_clVector);

	} catch(Error error) {
		std::cout << "Error in update()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}


void 
clParticles::calcHash() {
	try {
		kCalcHash.setArg(0, bPos);
		kCalcHash.setArg(1, bHash);
		kCalcHash.setArg(2, bIndex);
		kCalcHash.setArg(3, m_numParticles);
		kCalcHash.setArg(4, bParams);

		NDRange global(m_numParticles);
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kCalcHash, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in calcHash()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}

void 
clParticles::findCellBounds() {
	try {
		kFindCellBounds.setArg(0, bPos);
		kFindCellBounds.setArg(1, bVel);
		kFindCellBounds.setArg(2, bSortedPos);
		kFindCellBounds.setArg(3, bSortedVel);
		kFindCellBounds.setArg(4, bIndex);
		kFindCellBounds.setArg(5, bHash);
		kFindCellBounds.setArg(6, bCellStart);
		kFindCellBounds.setArg(7, bCellEnd);
		kFindCellBounds.setArg(8, m_numParticles);

		NDRange global(m_numParticles);
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kFindCellBounds, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in findCellBounds()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}

void 
clParticles::initializeDensity(float m_snow, float m_mixture) {
	try {
		kInitializeDensity.setArg(0, bPos);
		kInitializeDensity.setArg(1, bInitDen);
		kInitializeDensity.setArg(2, m_snow);
		kInitializeDensity.setArg(3, m_mixture);
		kInitializeDensity.setArg(4, bCellStart);
		kInitializeDensity.setArg(5, bCellEnd);
		kInitializeDensity.setArg(6, bIndex);
		kInitializeDensity.setArg(7, m_numParticles);
		kInitializeDensity.setArg(8, bParams);

		NDRange global(g_psystem->getNumParticles());
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kInitializeDensity, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in kInitializeDensity()! " << error.what() << "(" << error.err() << ")" << std::endl;
	}
}

void 
clParticles::computeDensity() {
try {
		kComputeDensity.setArg(0, bPos);
		kComputeDensity.setArg(1, bVel);
		kComputeDensity.setArg(2, bSimVar);
		kComputeDensity.setArg(3, bParVar);
		kComputeDensity.setArg(4, bCellStart);
		kComputeDensity.setArg(5, bCellEnd);
		kComputeDensity.setArg(6, bIndex);
		kComputeDensity.setArg(7, m_numParticles);
		kComputeDensity.setArg(8, bParams);

		NDRange global(g_psystem->getNumParticles());
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kComputeDensity, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in kComputeDensity()! " << error.what() << "(" << error.err() << ")" << std::endl;
	}
}

void 
clParticles::computePressure() {
try {
		kComputePressure.setArg(0, bPos);
		kComputePressure.setArg(1, bVel);
		kComputePressure.setArg(2, bSimVar);
		kComputePressure.setArg(3, bParVar);
		kComputePressure.setArg(4, bCellStart);
		kComputePressure.setArg(5, bCellEnd);
		kComputePressure.setArg(6, bIndex);
		kComputePressure.setArg(7, m_numParticles);
		kComputePressure.setArg(8, bParams);

		NDRange global(g_psystem->getNumParticles());
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kComputePressure, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in kComputePressure()! " << error.what() << "(" << error.err() << ")" << std::endl;
	}
}

void 
clParticles::computeForces() {
try {
		kComputeForces.setArg(0, bPos);
		kComputeForces.setArg(1, bVel);
		kComputeForces.setArg(2, bSimVar);
		kComputeForces.setArg(3, bParVar);
		kComputeForces.setArg(4, bForce);
		kComputeForces.setArg(5, bColor);
		kComputeForces.setArg(6, bCellStart);
		kComputeForces.setArg(7, bCellEnd);
		kComputeForces.setArg(8, bIndex);
		kComputeForces.setArg(9, TIME_STEP);
		kComputeForces.setArg(10, m_numParticles);
		kComputeForces.setArg(11, bParams);

		NDRange global(g_psystem->getNumParticles());
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kComputeForces, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in kComputeForces()! " << error.what() << "(" << error.err() << ")" << std::endl;
	}
}

void
clParticles::integrate() {
	try {
		kIntegrate.setArg(0, bPos);
		kIntegrate.setArg(1, bVel);
		kIntegrate.setArg(2, bSimVar);
		kIntegrate.setArg(3, bForce);
		kIntegrate.setArg(4, bIndex);
		kIntegrate.setArg(5, TIME_STEP);
		kIntegrate.setArg(6, m_numParticles);
		kIntegrate.setArg(7, bParams);

		NDRange global(m_numParticles);
		NDRange local(1);
		g_clAdmin->getCommandQueue().enqueueNDRangeKernel(kIntegrate, NullRange, global, local);
	} catch(Error error) {
		std::cout << "Error in integrate()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }

}

void
clParticles::reset(ParticleConfig config) {
	switch (config) {
		default:
		case CONFIG_RANDOM:
			{
				int p = 0, v = 0;
				for ( uint i = 0; i < m_numParticles; ++i ) {
					cPos[p++] = 2.f*RAND - 1.f;
					cPos[p++] = 2.f*RAND - 1.f;
					cPos[p++] = 2.f*RAND - 1.f;
					cPos[p++] = 1.0f;
					cVel[v++] = 0.0f;
					cVel[v++] = 0.0f;
					cVel[v++] = 0.0f;
					cVel[v++] = 0.0f;
				}
			}
			break;
		case CONFIG_GRID:
			{
				float jitter = m_params.particleRadius * 0.05f;
				uint s = (int)ceilf(powf((float)m_numParticles, 1.0f / 3.0f));
				float spacing = 2.0f * m_params.particleRadius;
				for ( uint l = 0; l < s; ++l )
					for ( uint k = 0; k < s; ++k )
						for ( uint j = 0; j < s; ++j ) {
							uint i = (l * s * s) + (k * s) + j;
							if ( i < m_numParticles ) {
								cPos[i*4] = (spacing * j) + m_params.particleRadius - 1.0f + (RAND * 2.0f - 1.0f) * jitter;
								cPos[i*4+1] = (spacing * k) + m_params.particleRadius - 1.0f + (RAND * 2.0f - 1.0f) * jitter;
								cPos[i*4+2] = (spacing * l) + m_params.particleRadius - 0.5f + (RAND * 2.0f - 1.0f) * jitter;
								cPos[i*4+3] = 1.0f;
								cVel[i*4] = 0.0f;
								cVel[i*4+1] = 0.0f;
								cVel[i*4+2] = 0.0f;
								cVel[i*4+3] = 0.0f;
							}
						}
			}
			break;
	}
	glBindBuffer(GL_ARRAY_BUFFER, m_posVBO);
	glBufferData(GL_ARRAY_BUFFER, m_numParticles * sizeof(Vector4), cPos, GL_DYNAMIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glFinish();
	copyDataToGPU(cVel, bVel, sizeof(Vector4) * m_numParticles);
	copyDataToGPU(&m_params, bParams, sizeof(simParams_t));


	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	float vf = 0.5f; // volume fraction
	
	float m_snow =	0.00006; // mass of snow.
	float m_air =	0.00002; // mass of air.
	
	float m_mixture = (m_snow * vf) + ((1-vf) * m_air); // mass of mixture fluid.

	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// INITIALIZE DENSITY VALUES AT PARTICLES
	g_clAdmin->getCommandQueue().enqueueAcquireGLObjects(&m_clVector);
	calcHash();
	sort();
	findCellBounds();
	initializeDensity(m_snow, m_mixture);
	g_clAdmin->getCommandQueue().enqueueReleaseGLObjects(&m_clVector);

	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	copyDataFromGPU(cInitDen, bInitDen, sizeof(Vector4) * m_numParticles);

	float den_snow = 0.0f; // density of snow
	float den_mixture = 0.0f; // density of mixture fluid
	for ( uint i = 0; i < m_numParticles; ++i ) { // determine max of den_snow and den_mixture of the particles
		if ( cInitDen[i*4] > den_snow )
			den_snow = cInitDen[i*4];
		if ( cInitDen[i*4+1] > den_mixture ) 
			den_mixture = cInitDen[i*4+1];
	}

	float den_air = ( den_mixture - vf*den_snow ) / ( 1 - vf ); // density of air 

	float kv_snow = 0.000411f; // kinematic viscosity snow
	float kv_air =	0.00011f;	// kinematic viscosity air

	float stiffness = 30000.0f; // stiffness parameter (k)

	float viscosity = 65.96f; // viscosity coefficient 

	float mu_not  = (kv_air*den_air*den_snow - kv_snow*den_snow*den_air) / (den_snow - den_air); 
	float kv = (kv_snow*den_snow - kv_air*den_air) / (den_snow - den_air); // mixture kinematic viscosity
	float dv = mu_not + (kv * den_mixture); // mixture dynamic viscosity

	m_params.stiffness = stiffness;
	m_params.mu0 = mu_not;
	m_params.kv = kv;
	m_params.dv = dv;
	m_params.viscosity = viscosity;

	printf("MASS SNOW: %f\n", m_snow);
	printf("MASS AIR: %f\n", m_air);
	printf("MASS MIXTURE: %f\n", m_mixture);
	printf("DENSITY SNOW: %f\n", den_snow);
	printf("DENSITY MIXTURE: %f\n", den_mixture);
	printf("DENSITY AIR: %f\n", den_air);
	printf("DYN VIS : %f\n", dv);


	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// SET VALUES FOR PARTICLES.
	for ( uint i = 0; i < m_numParticles; ++i ) {
		cSimVar[i*4] = den_mixture; // density of particles
		cSimVar[i*4+1] = den_mixture; // reference density
		cSimVar[i*4+2] = den_mixture; // rest density of fluid
		cSimVar[i*4+3] = 0.0f; // pressure 
		cParVar[i*4] = 0.0f;
		cParVar[i*4+1] = 0.0f;
		cParVar[i*4+2] = 0.0f;
		cParVar[i*4+3] = m_mixture;
	}



	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	

	copyDataToGPU(cSimVar, bSimVar, sizeof(Vector4) * m_numParticles);
	copyDataToGPU(cParVar, bParVar, sizeof(Vector4) * m_numParticles);
	
	copyDataToGPU(&m_params, bParams, sizeof(simParams_t));
}

void 
clParticles::copyDataToGPU(const void * cpuData, Buffer gpuData, ::size_t size) {
	try {
		g_clAdmin->getCommandQueue().enqueueWriteBuffer(gpuData, CL_TRUE, 0, size, cpuData);
	} catch(Error error) {
		std::cout << "Error in copyDataToGPU()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}

void 
clParticles::copyDataFromGPU(void * cpuData, Buffer gpuData, ::size_t size) {
	try {
		g_clAdmin->getCommandQueue().enqueueReadBuffer(gpuData, CL_TRUE, 0, size, cpuData);
	} catch(Error error) {
		std::cout << "Error in copyDataFromGPU()! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}
