#include "clSort.h"

using namespace cl;

clSort::clSort(const char* fileName) :
m_fileName(fileName)
{
	init();
}

clSort::~clSort(void)
{
}

void
clSort::init() {
	try {
		createProgram();

		kBitonicSortLocal = Kernel(m_clSortProgram, "bitonicSortLocal");
		kBitonicSortLocal1 = Kernel(m_clSortProgram, "bitonicSortLocal1");
		kBitonicMergeGlobal = Kernel(m_clSortProgram, "bitonicMergeGlobal");
		kBitonicMergeLocal = Kernel(m_clSortProgram, "bitonicMergeLocal");

	} catch(Error error) {
		std::cout << "Error! " << error.what() << "(" << error.err() << ")" << std::endl;
    }
}

static std::string readFromFile(const char* file) {
	std::ifstream sourceFile(file);
	std::string sourceCode(std::istreambuf_iterator<char>(sourceFile), (std::istreambuf_iterator<char>()));

	return sourceCode;
}

void
clSort::createProgram() {
	std::string sourceCode = readFromFile(m_fileName);
	//std::cout << sourceCode.c_str() << std::endl;
	Program::Sources source(1, std::make_pair(sourceCode.c_str(), sourceCode.length()+1));

	m_clSortProgram = Program(g_clAdmin->getContext(), source);
	m_clSortProgram.build(g_clAdmin->getDevices());
}

static cl_uint factorRadix2(cl_uint& log2L, cl_uint L){
    if(!L){
        log2L = 0;
        return 0;
    }else{
        for(log2L = 0; (L & 1) == 0; L >>= 1, log2L++);
        return L;
    }
}

//Note: logically shared with BitonicSort_b.cl!
static const unsigned int LOCAL_SIZE_LIMIT = 512U;

void
clSort::bitonicSort(
	Buffer m_dstKey,
    Buffer m_dstVal,
    Buffer m_srcKey,
    Buffer m_srcVal ) 
	{

	uint arrayLength = g_psystem->getNumParticles();
	uint batch = 1;
	uint dir = 0;

	CommandQueue cq = g_clAdmin->getCommandQueue();

	if(arrayLength < 2)
        return;

    //Only power-of-two array lengths are supported so far
    cl_uint log2L;
    cl_uint factorizationRemainder = factorRadix2(log2L, arrayLength);

    dir = (dir != 0);

    if (arrayLength <= LOCAL_SIZE_LIMIT)
    {
        //Launch bitonicSortLocal
		kBitonicSortLocal.setArg(0, m_dstKey);
		kBitonicSortLocal.setArg(1, m_dstVal);
		kBitonicSortLocal.setArg(2, m_srcKey);
		kBitonicSortLocal.setArg(3, m_srcVal);
		kBitonicSortLocal.setArg(4, arrayLength);
		kBitonicSortLocal.setArg(5, dir);

		NDRange global(arrayLength / 2);
		NDRange local(LOCAL_SIZE_LIMIT / 2);
		cq.enqueueNDRangeKernel(kBitonicSortLocal, NullRange, global, local);
		cq.finish();
    }
    else
    {
        //Launch bitonicSortLocal1
		kBitonicSortLocal1.setArg(0, m_dstKey);
		kBitonicSortLocal1.setArg(1, m_dstVal);
		kBitonicSortLocal1.setArg(2, m_srcKey);
		kBitonicSortLocal1.setArg(3, m_srcVal);

		NDRange global(arrayLength / 2);
		NDRange local(LOCAL_SIZE_LIMIT / 2);
		cq.enqueueNDRangeKernel(kBitonicSortLocal1, NullRange, global, local);
		cq.finish();

        for(unsigned int size = 2 * LOCAL_SIZE_LIMIT; size <= arrayLength; size <<= 1)
        {
            for(unsigned stride = size / 2; stride > 0; stride >>= 1)
            {
                if(stride >= LOCAL_SIZE_LIMIT)
                {
                    //Launch bitonicMergeGlobal
					kBitonicMergeGlobal.setArg(0, m_dstKey);
					kBitonicMergeGlobal.setArg(1, m_dstVal);
					kBitonicMergeGlobal.setArg(2, m_dstKey);
					kBitonicMergeGlobal.setArg(3, m_dstVal);
					kBitonicMergeGlobal.setArg(4, arrayLength);
					kBitonicMergeGlobal.setArg(5, size);
					kBitonicMergeGlobal.setArg(6, stride);
					kBitonicMergeGlobal.setArg(7, dir);

					NDRange global(arrayLength / 2);
					NDRange local(LOCAL_SIZE_LIMIT / 4);
					cq.enqueueNDRangeKernel(kBitonicMergeGlobal, NullRange, global, local);
					cq.finish();
                }
                else
                {
                    //Launch bitonicMergeLocal
					kBitonicMergeLocal.setArg(0, m_dstKey);
					kBitonicMergeLocal.setArg(1, m_dstVal);
					kBitonicMergeLocal.setArg(2, m_dstKey);
					kBitonicMergeLocal.setArg(3, m_dstVal);
					kBitonicMergeLocal.setArg(4, arrayLength);
					kBitonicMergeLocal.setArg(5, stride);
					kBitonicMergeLocal.setArg(6, size);
					kBitonicMergeLocal.setArg(7, dir);

					NDRange global(arrayLength / 2);
					NDRange local(LOCAL_SIZE_LIMIT / 2);
					cq.enqueueNDRangeKernel(kBitonicMergeLocal, NullRange, global, local);
					cq.finish();
                }
            }
        }
    }
}